from .deserialization import Deserializer
from .serialization import Serializer

_default_serializer = Serializer()
serialize = _default_serializer.serialize
serializer = _default_serializer.register()
serializer_of = _default_serializer.register
unregister_serializers = _default_serializer.unregister

_default_deserializer = Deserializer()
deserialize = _default_deserializer.deserialize
deserializer = _default_deserializer.register()
deserializer_of = _default_deserializer.register
unregister_deserializers = _default_deserializer.unregister



from functools import partial
from typing import Iterable, Dict, Callable, cast

from more_itertools.recipes import consume


def remove_type_data(data, type_key: str) -> None:
    callback = cast(
        Callable[[dict], None], partial(_delete_type_key, type_key=type_key)
    )
    _iterate_data(data, callback)


def rename_types(data, type_key: str, rename_map: Dict[str, str]) -> None:
    callback = cast(
        Callable[[dict], None],
        partial(_rename_types, type_key=type_key, rename_map=rename_map),
    )
    _iterate_data(data, callback)


def _iterate_data(data, callback: Callable[[dict], None]) -> None:
    if isinstance(data, dict):
        callback(data)
        consume(_iterate_data(v, callback) for v in data.values())
    elif isinstance(data, Iterable) and not isinstance(data, str):
        consume(_iterate_data(i, callback) for i in data)


def _delete_type_key(data: dict, type_key: str) -> None:
    data.pop(type_key, None)


def _rename_types(data: dict, type_key: str, rename_map: Dict[str, str]) -> None:
    if type_key not in data:
        return
    renamed = rename_map.get(data[type_key])
    if renamed is not None:
        data[type_key] = renamed


def _add_sub_parser_common_args(sub_parser):
    sub_parser.add_argument(
        "-i", "--infile", type=argparse.FileType("r"), required=True
    )
    sub_parser.add_argument(
        "-o", "--outfile", type=argparse.FileType("w"), required=True
    )


def _exec_remove_type_data(data, args):
    remove_type_data(data, args.type_key)


def _exec_rename_types(data, args):
    rename_map = dict(r.split(":") for r in args.rename_map)
    rename_types(data, args.type_key, rename_map)


if __name__ == "__main__":
    import argparse
    import json

    parser = argparse.ArgumentParser(
        description="Fix serialized data to reflect changes in the code."
    )
    sub_parsers = parser.add_subparsers()

    parser_del = sub_parsers.add_parser(
        "remove_types",
        description="Deletes the type key from the data, to allow deserialization based on type hints only.",
    )
    _add_sub_parser_common_args(parser_del)
    parser_del.add_argument("-t", "--type-key", type=str, required=True)
    parser_del.set_defaults(func=_exec_remove_type_data)

    parser_rename = sub_parsers.add_parser(
        "rename_types", description="Renames types to reflect changes in the code."
    )
    _add_sub_parser_common_args(parser_rename)
    parser_rename.add_argument("-t", "--type-key", type=str, required=True)
    parser_rename.add_argument(
        "-r",
        "--rename-map",
        nargs="+",
        help="One or more arguments of <old-name>:<new-name>",
    )
    parser_rename.set_defaults(func=_exec_rename_types)

    args = parser.parse_args()

    try:
        data = json.load(args.infile)
        args.func(data, args)
        json.dump(data, args.outfile)
    finally:
        args.infile.close()
        args.outfile.flush()
        args.outfile.close()



from datetime import datetime

from yasoo.utils import type_to_string, fully_qualified_string_to_type


def serialize_datetime(d: datetime) -> dict:
    return {"time": d.timestamp()}


def deserialize_datetime(d: dict) -> datetime:
    return datetime.fromtimestamp(d["time"])


def serialize_type(obj: type) -> dict:
    return {"fully_qualified_name": type_to_string(obj, fully_qualified=True)}


def deserialize_type(data: dict, _) -> type:
    return fully_qualified_string_to_type(data["fully_qualified_name"])



from functools import lru_cache
from importlib import import_module
from typing import Dict, Any, Union, List, Optional, Tuple

import attr
from attr.exceptions import NotAnAttrsClassError

try:
    import dataclasses
except ModuleNotFoundError:
    dataclasses = None

try:
    # Python 3.6
    from typing import _ForwardRef as ForwardRef, GenericMeta as GenericType

    def _get_origin(t: GenericType):
        return t.__extra__ or t.__origin__

    def _is_optional(t: type):
        return (
            getattr(t, "__origin__", None) is Union
            and len(t.__args__) == 2
            and any(a is NoneType for a in t.__args__)
        )


except ImportError:
    # Python >=3.7
    from typing import ForwardRef, _GenericAlias as GenericType

    def _get_origin(t: GenericType):
        return t.__origin__

    def _is_optional(t: type):
        return (
            isinstance(t, GenericType)
            and _get_origin(t) is Union
            and len(t.__args__) == 2
            and any(a is NoneType for a in t.__args__)
        )


generic_types = (GenericType,)

try:
    # Python >= 3.9
    from types import GenericAlias

    generic_types += (GenericAlias,)
except ImportError:
    pass


NoneType = type(None)
SUPPORTED_PRIMITIVES = {bool, int, float, str}
_SUPPORTED_PRIMITIVES = tuple(SUPPORTED_PRIMITIVES)


@attr.attrs
class Field:
    name: str = attr.attrib()
    field_type: type = attr.attrib()
    mandatory: bool = attr.attrib()
    validator: Optional[callable] = attr.attrib(default=None)
    converter: Optional[callable] = attr.attrib(default=None)


def resolve_types(
    to_resolve: Dict[Union[type, str], Any], globals: Dict[str, Any]
) -> Dict[type, Any]:
    return {_resolve_type(globals, k): v for k, v in to_resolve.items()}


@lru_cache(None)
def get_fields(obj_type: type) -> List[Field]:
    try:
        return [
            Field(f.name, f.type, f.default == attr.NOTHING, f.validator, f.converter)
            for f in attr.fields(obj_type)
        ]
    except NotAnAttrsClassError:
        try:
            return [
                Field(f.name, f.type, _dataclass_field_mandatory(f))
                for f in dataclasses.fields(obj_type)
            ]
        except (TypeError, AttributeError):
            pass
    raise TypeError("can only serialize attrs or dataclass classes")


def normalize_method(method) -> callable:
    return method.__func__ if isinstance(method, staticmethod) else method


def normalize_type(
    t: Union[type, GenericType], all_globals: Optional[dict] = None
) -> Tuple[type, tuple]:
    all_globals = all_globals or {}
    if t == Any:
        t = None
    if _is_optional(t):
        return normalize_type(
            next(a for a in t.__args__ if a is not NoneType), all_globals
        )
    if isinstance(t, ForwardRef):
        t = t.__forward_arg__
    if isinstance(t, str) and t in all_globals:
        return all_globals[t], tuple()
    return _normalize_type(t)


@lru_cache(None)
def _normalize_type(t: Union[type, GenericType]) -> Tuple[type, tuple]:
    if isinstance(t, generic_types):
        real_type = _get_origin(t)
        generic_args = t.__args__
    elif t is None or isinstance(t, type):
        real_type = t
        generic_args = tuple()
    else:
        raise TypeError(
            f"Found type annotation {t}, which is not a type and not a generic."
        )
    return real_type, generic_args


def is_obj_supported_primitive(obj):
    return isinstance(obj, _SUPPORTED_PRIMITIVES) or obj is None


@lru_cache(None)
def type_to_string(t: type, fully_qualified: bool) -> str:
    name = t.__name__
    if fully_qualified:
        return ".".join((t.__module__, name))
    else:
        return name


@lru_cache(None)
def fully_qualified_string_to_type(fully_qualified_type_name: str) -> type:
    module_name = fully_qualified_type_name[: fully_qualified_type_name.rindex(".")]
    class_name = fully_qualified_type_name[len(module_name) + 1 :]
    return getattr(import_module(module_name), class_name)


def _resolve_type(globals, t):
    return globals.get(t) if isinstance(t, str) else t


def _dataclass_field_mandatory(field):
    return (
        field.default == dataclasses.MISSING
        and field.default_factory == dataclasses.MISSING
    )



from typing import TypeVar, Type, Callable, overload, List, Set, Tuple, Dict

K = TypeVar("K")
V = TypeVar("V")


class _List:
    @overload
    def __getitem__(self, item: Type[V]) -> Callable[[], List[V]]:
        ...

    def __getitem__(self, item: Callable[[], V]) -> Callable[[], List[V]]:
        return List[item]


class _Set:
    @overload
    def __getitem__(self, item: Type[V]) -> Callable[[], Set[V]]:
        ...

    def __getitem__(self, item: Callable[[], V]) -> Callable[[], Set[V]]:
        return Set[item]


class _Dict:
    @overload
    def __getitem__(self, item: Tuple[Type[K], Type[V]]) -> Callable[[], Dict[K, V]]:
        ...

    @overload
    def __getitem__(
        self, item: Tuple[Type[K], Callable[[], V]]
    ) -> Callable[[], Dict[K, V]]:
        ...

    @overload
    def __getitem__(
        self, item: Tuple[Callable[[], K], Type[V]]
    ) -> Callable[[], Dict[K, V]]:
        ...

    def __getitem__(
        self, item: Tuple[Callable[[], K], Callable[[], V]]
    ) -> Callable[[], Dict[K, V]]:
        return Dict[item[0], item[1]]


List_ = _List()
Set_ = _Set()
Dict_ = _Dict()



import json
import warnings
from contextlib import contextmanager
from datetime import datetime
from enum import Enum
from inspect import signature
from typing import Dict, Any, Union, Mapping, Iterable, Callable, Optional

from yasoo.default_customs import serialize_type, serialize_datetime
from yasoo.objects import DictWithSerializedKeys
from yasoo.utils import normalize_type, type_to_string
from .constants import ENUM_VALUE_KEY, ITERABLE_VALUE_KEY
from .utils import (
    resolve_types,
    get_fields,
    normalize_method,
    Field,
    is_obj_supported_primitive,
)


class Serializer:
    def __init__(self) -> None:
        super().__init__()
        self._custom_serializers: Dict[type, Callable[[Any], Dict[str, Any]]] = {
            datetime: serialize_datetime,
        }
        self._inheritance_serializers: Dict[type, Callable[[Any], Dict[str, Any]]] = {
            type: serialize_type,
        }

    def register(
        self, type_to_register: Optional[type] = None, include_descendants: bool = False
    ):
        """
        Registers a custom serialization method that takes the object to be serialized and returns a json-serializable
        dictionary.

        :param type_to_register: The type of objects this method serializes. Can be a string, but then ``serialize``
            should be called with the ``globals`` parameter.
        :param include_descendants: Whether to use the registered method for objects inheriting from the given type
            or only for objects of the given type itself.
        """

        def registration_method(
            serialization_method: Union[Callable[[Any], Dict[str, Any]], staticmethod]
        ):
            method = normalize_method(serialization_method)
            t = type_to_register
            if t is None:
                t = next(iter(signature(method).parameters.values())).annotation
            self._custom_serializers[t] = method
            if include_descendants:
                self._inheritance_serializers[t] = method
            return serialization_method

        return registration_method

    @contextmanager
    def unregister(self, *types: type):
        """
        Temporarily unregisters registered serializers, so ``serialize`` will use the default serialization
        algorithm.

        :param types: The types to serialize using the default algorithm.
        :return:
        """
        types_funcs = [
            (type_, self._custom_serializers.pop(type_, None)) for type_ in types
        ]
        try:
            yield
        finally:
            for type_, func in types_funcs:
                if func is not None:
                    self._custom_serializers[type_] = func
        pass

    def serialize(
        self,
        obj,
        type_key: Optional[str] = "__type",
        fully_qualified_types: bool = True,
        preserve_iterable_types: bool = False,
        stringify_dict_keys: bool = True,
        globals: Optional[Dict[str, Any]] = None,
    ) -> Optional[Union[bool, int, float, str, list, Dict[str, Any]]]:
        """
        Serializes an object to a json-serializable dictionary or list,
        or returns the object itself if it's json-serializable.

        :param obj: The object to serialize.
        :param type_key: The key in the resulting dictionary to contain the type name for non-primitive objects.
            Can be ``None`` to omit this key and rely on type hints when deserializing.
        :param fully_qualified_types: Whether to use fully qualified type names in the ``type_key``,
            i.e. whether to write "my_package.MyType" or just "MyType".
        :param preserve_iterable_types: Whether to serialize iterables as dictionaries with their type
            under ``type_key``, so they will be deserialized back to their type and not as a list.
        :param stringify_dict_keys: Whether to convert non-string dictionary keys to strings in a way that will
            allow deserializing them back to their original type. Useful for json.
        :param globals: If custom serialization methods were registered and used forward reference
            ('Foo' instead of Foo), this parameter should be a dictionary from type name to type, most easily
            acquired using the built-in ``globals()`` function.
        """
        if is_obj_supported_primitive(obj):
            return obj

        if globals:
            self._custom_serializers = resolve_types(self._custom_serializers, globals)

        result = self._serialize(
            obj,
            type_key,
            fully_qualified_types,
            preserve_iterable_types,
            stringify_dict_keys,
            inner=False,
        )
        result = _convert_to_json_serializable(result)
        return result

    def _serialize(
        self,
        obj,
        type_key,
        fully_qualified_types,
        preserve_iterable_types,
        stringify_dict_keys,
        inner=True,
    ):
        serialization_method = self._custom_serializers.get(type(obj))
        if serialization_method is None:
            for base_class, method in self._inheritance_serializers.items():
                if isinstance(obj, base_class):
                    serialization_method = method
                    break
        if serialization_method is not None:
            result = serialization_method(obj)
        else:
            try:
                result = self._serialize_data_class(
                    obj,
                    type_key,
                    fully_qualified_types,
                    preserve_iterable_types,
                    stringify_dict_keys,
                )
            except TypeError:
                if isinstance(obj, Enum):
                    result = {ENUM_VALUE_KEY: obj.value}
                elif isinstance(obj, Mapping):
                    result = self._serialize_mapping(
                        obj,
                        type_key,
                        fully_qualified_types,
                        preserve_iterable_types,
                        stringify_dict_keys,
                    )
                elif isinstance(obj, Iterable) and not isinstance(obj, str):
                    serialized = self._serialize_iterable(
                        obj,
                        type_key,
                        fully_qualified_types,
                        preserve_iterable_types,
                        stringify_dict_keys,
                    )
                    if isinstance(obj, list) or not preserve_iterable_types:
                        return serialized
                    result = {ITERABLE_VALUE_KEY: serialized}
                elif not inner:
                    raise
                else:
                    return obj

        if type_key is not None and type_key not in result:
            result[type_key] = self._get_type_data(obj, fully_qualified_types)
        return result

    def _serialize_data_class(
        self,
        obj,
        type_key,
        fully_qualified_types,
        preserve_iterable_types,
        stringify_dict_keys,
    ):
        fields = get_fields(type(obj))
        result = {
            f.name: self._serialize(
                getattr(obj, f.name),
                type_key,
                fully_qualified_types,
                preserve_iterable_types,
                stringify_dict_keys,
            )
            for f in fields
        }
        self._warn_for_possible_problems_in_deserialization(
            obj, fields, result, type_key is not None
        )
        return result

    def _serialize_iterable(
        self,
        obj: Iterable,
        type_key,
        fully_qualified_types,
        preserve_iterable_types,
        stringify_dict_keys,
    ):
        return [
            self._serialize(
                item,
                type_key,
                fully_qualified_types,
                preserve_iterable_types,
                stringify_dict_keys,
            )
            for item in obj
        ]

    def _serialize_mapping(
        self,
        obj: Mapping,
        type_key,
        fully_qualified_types,
        preserve_iterable_types,
        stringify_dict_keys,
    ):
        result = self._serialize_mapping_values(
            obj,
            type_key=type_key,
            fully_qualified_types=fully_qualified_types,
            preserve_iterable_types=preserve_iterable_types,
            stringify_dict_keys=stringify_dict_keys,
        )
        if any(not self._key_ok(k, stringify_dict_keys) for k in result.keys()):
            obj_type = self._get_type_data(obj, fully_qualified_types)
            d = self._serialize_complex_keys(
                result, obj_type, type_key, fully_qualified_types
            )
            result = self._serialize(
                d,
                type_key=type_key,
                fully_qualified_types=fully_qualified_types,
                preserve_iterable_types=preserve_iterable_types,
                stringify_dict_keys=stringify_dict_keys,
            )
        return result

    def _serialize_complex_keys(
        self, obj: dict, obj_type: str, type_key, fully_qualified_types
    ):
        def serialize_key(k):
            return json.dumps(
                self.serialize(
                    k,
                    type_key=type_key,
                    fully_qualified_types=fully_qualified_types,
                    preserve_iterable_types=True,
                )
            )

        try:
            data = {serialize_key(k): v for k, v in obj.items()}
            return DictWithSerializedKeys(data, obj_type)
        except TypeError:
            raise ValueError(
                f"Mapping {obj} contains a key which is not json-serializable and not yasoo-serializable"
            )

    def _serialize_mapping_values(
        self,
        obj,
        type_key,
        fully_qualified_types,
        preserve_iterable_types,
        stringify_dict_keys,
    ):
        return {
            k: self._serialize(
                v,
                type_key,
                fully_qualified_types,
                preserve_iterable_types,
                stringify_dict_keys,
            )
            for k, v in obj.items()
        }

    @classmethod
    def _key_ok(cls, key, stringify_dict_keys):
        return (
            isinstance(key, str)
            if stringify_dict_keys
            else is_obj_supported_primitive(key)
        )

    @classmethod
    def _warn_for_possible_problems_in_deserialization(
        cls,
        obj,
        fields: Iterable[Field],
        data: Dict[str, Any],
        type_key_present: bool,
    ) -> None:
        for f in fields:
            value = data[f.name]
            if not type_key_present:
                cls._check_for_unknown_dicts(f, value, obj.__class__.__name__)
            cls._check_for_unconvertables_or_invalid(
                obj, f, value, obj.__class__.__name__
            )

    @classmethod
    def _check_for_unknown_dicts(cls, f, value, obj_class_name):
        try:
            real_type, generic_args = normalize_type(f.field_type)
        except TypeError:
            real_type = generic_args = None

        if isinstance(value, list) and value:
            value = value[0]
            real_type = generic_args[0] if generic_args else None

        if not isinstance(value, dict):
            return

        if real_type is None:
            cls._warn(
                f'Field "{f.name}" in obj "{obj_class_name}" is a dict or an instance and has no type hint'
            )
        else:
            try:
                if not issubclass(real_type, Mapping) and not issubclass(
                    real_type, Iterable
                ):
                    get_fields(real_type)
            except TypeError:
                cls._warn(
                    f'Field "{f.name}" in obj "{obj_class_name}" is a dict or an instance but its type hint is an unsupported class. Make sure you register a deserializer'
                )

    @classmethod
    def _check_for_unconvertables_or_invalid(cls, obj, f, value, obj_class_name):
        if f.converter is not None and not isinstance(value, dict):
            try:
                value = f.converter(value)
            except:
                cls._warn(
                    f'Field "{f.name}" in obj "{obj_class_name}" has value {value} that could not be converted using its converter'
                )
                return

        if f.validator is not None and not isinstance(value, dict):
            try:
                f.validator(obj, f, value)
            except:
                cls._warn(
                    f'Field "{f.name}" in obj "{obj.__class__.__name__}" has value {value} that doesn\'t match this field\'s validator'
                )
                return

        if f.converter is not None:
            cls._warn(
                f'Field "{f.name}" in obj "{obj.__class__.__name__}" has a converter'
            )

    @staticmethod
    def _get_type_data(obj, fully_qualified_types) -> str:
        return type_to_string(type(obj), fully_qualified_types)

    @staticmethod
    def _warn(warning):
        warnings.warn(warning, RuntimeWarning, stacklevel=2)


def _convert_to_json_serializable(obj) -> Union[int, float, str, list, dict, None]:
    if is_obj_supported_primitive(obj):
        return obj
    if isinstance(obj, Mapping):
        return {key: _convert_to_json_serializable(value) for key, value in obj.items()}
    if isinstance(obj, Iterable):
        return [_convert_to_json_serializable(item) for item in obj]
    raise TypeError(
        f'Found object of type "{type(obj).__name__}" which cannot be serialized'
    )



import json
from contextlib import contextmanager
from datetime import datetime
from enum import Enum
from inspect import signature
from itertools import zip_longest
from typing import (
    Optional,
    Type,
    Union,
    Callable,
    Dict,
    Any,
    TypeVar,
    Mapping,
    Iterable,
    List,
    Tuple,
    overload,
)

from yasoo.default_customs import deserialize_type, deserialize_datetime
from yasoo.utils import fully_qualified_string_to_type, NoneType
from .constants import ENUM_VALUE_KEY, ITERABLE_VALUE_KEY
from .objects import DictWithSerializedKeys
from .utils import (
    resolve_types,
    get_fields,
    normalize_method,
    normalize_type,
    is_obj_supported_primitive,
    SUPPORTED_PRIMITIVES,
)

T = TypeVar("T")


class Deserializer:
    def __init__(self) -> None:
        super().__init__()
        t = Dict[type, Callable[[Dict[str, Any], Type[T]], T]]
        self._custom_deserializers: Dict[Type[T], Callable[[Dict[str, Any]], T]] = {
            datetime: deserialize_datetime,
        }
        self._inheritance_deserializers: t = {
            type: deserialize_type,
        }

    def register(
        self,
        type_to_register: Optional[Union[Type[T], str]] = None,
        include_descendants: bool = False,
    ):
        """
        Registers a custom deserialization method that takes a dictionary and returns an instance of the registered
        type.

        :param type_to_register: The type of objects this method deserializes to. Can be a string, but then
            ``deserialize`` should be called with the ``globals`` parameter.
        :param include_descendants: Whether to use the registered method for objects inheriting from the given type
            or only for objects of the given type itself.
        """

        def registration_method(
            deserialization_method: Union[Callable[[Dict[str, Any]], T], staticmethod]
        ):
            method = normalize_method(deserialization_method)
            t = type_to_register
            if t is None:
                t = signature(method).return_annotation
            self._custom_deserializers[t] = method
            if include_descendants:
                self._inheritance_deserializers[t] = method
            return deserialization_method

        return registration_method

    @contextmanager
    def unregister(self, *types: Type[T]):
        """
        Temporarily unregisters registered deserializers, so ``deserialize`` will use the default deserialization
        algorithm.

        :param types: The types to deserialize using the default algorithm.
        :return:
        """
        types_funcs = [
            (type_, self._custom_deserializers.pop(type_, None)) for type_ in types
        ]
        try:
            yield
        finally:
            for type_, func in types_funcs:
                if func is not None:
                    self._custom_deserializers[type_] = func

    @overload
    def deserialize(
        self,
        data: Optional[Union[bool, int, float, str, list, Dict[str, Any]]],
        obj_type: Callable[[], T],
        type_key: Optional[str] = "__type",
        allow_extra_fields: bool = False,
        globals: Optional[Dict[str, Any]] = None,
    ) -> T:
        ...

    def deserialize(
        self,
        data: Optional[Union[bool, int, float, str, list, Dict[str, Any]]],
        obj_type: Optional[Type[T]] = None,
        type_key: Optional[str] = "__type",
        allow_extra_fields: bool = False,
        globals: Optional[Dict[str, Any]] = None,
    ) -> T:
        """
        Deserializes an object from a dictionary or a list of dictionaries,
        or returns the object itself if it's a primitive (including ``None``).

        :param data: The dictionary.
        :param obj_type: The type of the object to deserialize. Can only be ``None`` if ``data`` contains a type key.
        :param type_key: The key in ``data`` that contains the type name for non-primitive objects.
            Can be ``None`` if this key was omitted during serialization and deserialization should rely on type hints.
        :param allow_extra_fields: Whether to throw an exception if the data contains fields that are not in the type
            definition, or just ignore them.
        :param globals: A dictionary from type name to type, most easily acquired using the built-in ``globals()``
            function.
        """
        if globals:
            self._custom_deserializers = resolve_types(
                self._custom_deserializers, globals
            )

        return self._deserialize(
            data, obj_type, type_key, allow_extra_fields, globals or {}
        )

    def _deserialize(
        self,
        data: Optional[Union[bool, int, float, str, list, Dict[str, Any]]],
        obj_type: Optional[Type[T]],
        type_key: Optional[str],
        allow_extra_fields: bool,
        external_globals: Dict[str, Any],
    ):
        all_globals = dict(globals())
        all_globals.update(external_globals)
        if is_obj_supported_primitive(data):
            return data
        if isinstance(data, list):
            list_types = self._get_list_types(obj_type, data)
            return [
                self._deserialize(d, t, type_key, allow_extra_fields, all_globals)
                for t, d in list_types
            ]

        obj_type = self._get_object_type(obj_type, data, type_key, all_globals)
        if type_key in data:
            data.pop(type_key)
        real_type, generic_args = normalize_type(obj_type, all_globals)

        deserialization_method = self._custom_deserializers.get(
            obj_type, self._custom_deserializers.get(real_type)
        )
        if deserialization_method:
            return deserialization_method(data)
        for base_class, method in self._inheritance_deserializers.items():
            if issubclass(real_type, base_class):
                return method(data, real_type)

        key_type = None
        try:
            fields = {f.name: f for f in get_fields(obj_type)}
        except TypeError:
            if issubclass(real_type, Enum):
                return obj_type(data[ENUM_VALUE_KEY])
            elif issubclass(real_type, Mapping):
                key_type = generic_args[0] if generic_args else None
                if self._is_mapping_dict_with_serialized_keys(key_type, data):
                    obj_type = DictWithSerializedKeys
                    fields = {f.name: f for f in get_fields(obj_type)}
                    value_type = generic_args[1] if generic_args else Any
                    fields["data"].field_type = Dict[str, value_type]
                else:
                    return self._load_mapping(
                        data,
                        real_type,
                        generic_args,
                        type_key,
                        allow_extra_fields,
                        all_globals,
                    )
            elif issubclass(real_type, Iterable):
                # If we got here it means data is not a list, so obj_type came from the data itself and is safe to use
                return self._load_iterable(
                    data, obj_type, type_key, allow_extra_fields, all_globals
                )
            elif real_type != obj_type:
                return self._deserialize(
                    data, real_type, type_key, allow_extra_fields, external_globals
                )
            else:
                raise

        self._check_for_missing_fields(data, fields, obj_type)
        self._check_for_extraneous_fields(data, fields, obj_type, allow_extra_fields)
        self._load_inner_fields(data, fields, type_key, allow_extra_fields, all_globals)
        if obj_type is DictWithSerializedKeys:
            return self._load_dict_with_serialized_keys(
                obj_type(**data), key_type, type_key, allow_extra_fields, all_globals
            )
        return obj_type(**data)

    def _load_dict_with_serialized_keys(
        self,
        obj: DictWithSerializedKeys,
        key_type,
        type_key,
        allow_extra_fields,
        all_globals,
    ):
        data = {
            self._deserialize(
                json.loads(k), key_type, type_key, allow_extra_fields, all_globals
            ): v
            for k, v in obj.data.items()
        }
        obj_type = Deserializer._get_type(obj.original_type, all_globals)
        return obj_type(data)

    def _load_mapping(
        self,
        data: Mapping,
        obj_type,
        generic_args,
        type_key,
        allow_extra_fields,
        all_globals,
    ):
        val_type = generic_args[1] if len(generic_args) > 1 else None
        return obj_type(
            {
                k: self._deserialize(
                    v, val_type, type_key, allow_extra_fields, all_globals
                )
                for k, v in data.items()
            }
        )

    def _load_iterable(self, data, obj_type, type_key, allow_extra_fields, all_globals):
        return obj_type(
            self._deserialize(i, None, type_key, allow_extra_fields, all_globals)
            for i in data[ITERABLE_VALUE_KEY]
        )

    def _load_inner_fields(
        self, data, fields, type_key, allow_extra_fields, all_globals
    ):
        for key, value in data.items():
            field = fields[key]
            data[key] = self._deserialize(
                value, field.field_type, type_key, allow_extra_fields, all_globals
            )

    @classmethod
    def _is_mapping_dict_with_serialized_keys(cls, key_type, data):
        if (
            key_type
            and key_type not in SUPPORTED_PRIMITIVES
            and key_type is not NoneType
        ):
            return True
        if key_type is str:
            return False

        fields = {f.name: f for f in get_fields(DictWithSerializedKeys)}
        try:
            cls._check_for_missing_fields(data, fields, DictWithSerializedKeys)
        except ValueError:
            return False
        return True

    @staticmethod
    def _check_for_missing_fields(data, fields, obj_type):
        missing = {
            name
            for name, field in fields.items()
            if name not in data and field.mandatory
        }
        if missing:
            missing_str = '", "'.join(missing)
            raise ValueError(
                f'Missing fields "{missing_str}" for object type "{obj_type.__name__}". Data is:\n{json.dumps(data)}'
            )

    @staticmethod
    def _check_for_extraneous_fields(data, fields, obj_type, allow_extra_fields):
        extraneous = set(data.keys()).difference(fields)
        if extraneous and not allow_extra_fields:
            extraneous_str = '", "'.join(extraneous)
            raise ValueError(
                f'Found extraneous fields "{extraneous_str}" for object type "{obj_type.__name__}".'
                f"Data is:\n{json.dumps(data)}"
            )
        for e in extraneous:
            data.pop(e)

    @staticmethod
    def _get_list_types(
        type_hint: Optional[type], data: list
    ) -> List[Tuple[Optional[type], Any]]:
        if type_hint is None:
            return [(None, item) for item in data]
        _, generic_args = normalize_type(type_hint)
        if len(generic_args) == 1 or (
            len(generic_args) == 2 and generic_args[1] is ...
        ):
            return [(generic_args[0], item) for item in data]
        if len(generic_args) >= len(data):
            return list(zip(generic_args, data))
        return list(zip_longest(generic_args, data))

    @staticmethod
    def _get_object_type(
        obj_type: Optional[Type[T]],
        data: Dict[str, Any],
        type_key: str,
        all_globals: Dict[str, Any],
    ) -> type:
        if type_key in data:
            return Deserializer._get_type(data[type_key], all_globals)
        if obj_type is None or obj_type is Any:
            raise ValueError(
                f"type key not found in data and obj type could not be inferred.\nData: {json.dumps(data)}"
            )
        return obj_type

    @staticmethod
    def _get_type(type_name: str, all_globals: Dict[str, Any]) -> type:
        if "." not in type_name:
            return Deserializer._get_non_fully_qualified_type(type_name, all_globals)
        return fully_qualified_string_to_type(type_name)

    @staticmethod
    def _get_non_fully_qualified_type(
        type_name: str, all_globals: Dict[str, Any]
    ) -> type:
        if type_name == "list":
            return list
        if type_name == "set":
            return set
        if type_name == "tuple":
            return tuple
        if type_name == "dict":
            return dict
        if type_name not in all_globals:
            raise ValueError(f"type {type_name} not found in globals.")
        return all_globals[type_name]
